# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""Build IDF geometry from minimal inputs."""
from typing import Any, Dict, List, Tuple, Union  # noqa

from .geom.polygons import Polygon3D
from .geom.segments import Segment  # noqa
from .geom.vectors import Vector3D


class Zone(object):
    """Represents a single zone for translation into an IDF."""

    def __init__(self, name, surfaces):
        # type: (str, Dict[str, Any]) -> None
        """Initialise the zone object.

        :param name: A name for the zone.
        :param surfaces: The surfaces that make up the zone.

        """
        self.name = name
        self.walls = [s for s in surfaces['walls'] if s.area > 0]
        self.floors = surfaces['floors']
        self.roofs = surfaces['roofs']
        self.ceilings = surfaces['ceilings']


class Block(object):
    def __init__(self,
                 name,  # type: str
                 # type: Union[List[Tuple[float, float]], List[Tuple[int,
                 # int]]]
                 coordinates,
                 height,  # type: float
                 num_stories=1,  # type: int
                 below_ground_stories=0,  # type: int
                 below_ground_storey_height=2.5  # type: float
                 ):
        # type: (...) -> None
        """Represents a single block for translation into an IDF.

        :param name: A name for the block.
        :param coordinates: A list of (x, y) tuples representing the building outline.
        :param height: The height of the block roof above ground level.
        :param num_stories: The total number of stories including basement stories. Default : 1.
        :param below_ground_stories: The number of stories below ground. Default : 0.
        :param below_ground_storey_height: The height of each basement storey. Default : 2.5.

        """
        self.name = name
        if coordinates[0] == coordinates[-1]:
            coordinates.pop()
        self.coordinates = coordinates
        self.height = height
        self.num_stories = num_stories
        self.num_below_ground_stories = below_ground_stories
        self.below_ground_storey_height = below_ground_storey_height

    @property
    def stories(self):
        # type: () -> List[Dict[str, Any]]
        """A list of dicts of the surfaces of each storey in the block.

        :returns: list of dicts

        Example dict format::
            {'floors': [...],
             'ceilings': [...],
             'walls': [...],
             'roofs': [...],
             }

        """
        stories = []
        if self.num_below_ground_stories != 0:
            floor_no = -self.num_below_ground_stories
        else:
            floor_no = 0
        for floor, ceiling, wall, roof in zip(
                self.floors, self.ceilings, self.walls, self.roofs):
            stories.append({
                'storey_no': floor_no,
                'floors': floor, 'ceilings': ceiling, 'walls': wall, 'roofs': roof})
            floor_no += 1
        return stories

    @property
    def footprint(self):
        # type: () -> Polygon3D
        """Ground level outline of the block.

        :returns: A 2D outline of the block.

        """
        coordinates = [(v[0], v[1], 0) for v in self.coordinates]
        return Polygon3D(coordinates)

    @property
    def storey_height(self):
        # type: () -> float
        """Height of above ground stories.

        :returns: Average storey height.

        """
        return self.height / (self.num_stories - self.num_below_ground_stories)

    @property
    def floor_heights(self):
        # type: () -> List[float]
        """Floor height for each storey in the block.

        :returns: A list of floor heights.

        """
        lfl = self.lowest_floor_level
        sh = self.storey_height
        floor_heights = [lfl + sh * i for i in range(self.num_stories)]
        return floor_heights

    @property
    def ceiling_heights(self):
        # type: () -> List[float]
        """Ceiling height for each storey in the block.

        :returns: A list of ceiling heights.

        """
        lfl = self.lowest_floor_level
        sh = self.storey_height
        ceiling_heights = [lfl + sh * (i + 1) for i in range(self.num_stories)]
        return ceiling_heights

    @property
    def lowest_floor_level(self):
        # type: () -> float
        """Floor level of the lowest basement storey.

        :returns: Lowest floor height.

        """
        return -(self.num_below_ground_stories *
                 self.below_ground_storey_height)

    @property
    def walls(self):
        # type: () -> List[List[Polygon3D]]
        """Coordinates for each wall in the block.

        These are ordered as a list of lists, one for each storey.

        :returns: Coordinates for all walls.

        """
        walls = []
        for fh, ch in zip(self.floor_heights, self.ceiling_heights):
            floor_walls = [make_wall(edge, fh, ch)
                           for edge in self.footprint.edges]
            walls.append(floor_walls)
        return walls

    @property
    def floors(self):
        # type: () -> List[List[Polygon3D]]
        """Coordinates for each floor in the block.

        :returns: Coordinates for all floors.

        """
        floors = [[self.footprint.invert_orientation() + Vector3D(0, 0, fh)]
                  for fh in self.floor_heights]
        return floors

    @property
    def ceilings(self):
        # type: () -> List[Union[List[Polygon3D], str]]
        """Coordinates for each ceiling in the block.

        :returns: Coordinates for all ceilings.

        """
        ceilings = [[self.footprint + Vector3D(0, 0, ch)]
                    for ch in self.ceiling_heights[:-1]]

        ceilings.append('')
        return ceilings

    @property
    def roofs(self):
        # type: () -> List[List[Polygon3D]]
        """Coordinates for each roof of the block.

        This returns a list with an entry for each floor for consistency with
        the other properties of the Block object, but should only have roof
        coordinates in the list in the final position.

        :returns: Coordinates for all roofs.

        """
        roofs = [[None] for ch in self.ceiling_heights[:-1]]
        roofs.append([self.footprint + Vector3D(0, 0, self.height)])
        return roofs

    @property
    def surfaces(self):
        # type: () -> Dict[str, Any]
        """Coordinates for all the surfaces in the block.

        :returns: Coordinates for all surfaces.

        """
        return {'walls': self.walls,
                'ceilings': self.ceilings,
                'roofs': self.roofs,
                'floors': self.floors}


def make_wall(edge, floor_height, ceiling_height):
    # type: (Segment, float, float) -> Polygon3D
    """Create a polygon representing the vertices of a wall.

    :param edge: Segment of a floor outline at ground level.
    :param floor_height: Floor height.
    :param ceiling_height: Ceiling height.

    """
    return Polygon3D([edge.p1 + (0, 0, ceiling_height),  # upper left
                      edge.p1 + (0, 0, floor_height),  # lower left
                      edge.p2 + (0, 0, floor_height),  # lower right
                      edge.p2 + (0, 0, ceiling_height),  # upper right
                      ])



# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""pytest for polygons.py"""

from geomeppy.geom.polygons import (
    break_polygons, difference_3D_polys, intersect_3D_polys, Polygon, Polygon3D, union_2D_polys, union_3D_polys,
    Vector2D, Vector3D,
)
from geomeppy.geom.segments import Segment
from geomeppy.utilities import almostequal


def test_polygon_repr():
    # type: () -> None
    s2D = Polygon([(0,0), (2,0), (2,0), (0,0)])  # vertical
    assert eval(repr(s2D)) == s2D
    
    s3D = Polygon3D([(0,0,0), (2,0,0), (2,2,0), (0,2,0)])  # vertical
    assert eval(repr(s3D)) == s3D


def test_equal_polygon():
    # type: () -> None
    poly1 = Polygon([(1,0),(0,0),(0,1),(1,1)])
    poly2 = Polygon([(1,1),(1,0),(0,0),(0,1)])
    assert poly1 == poly2
    
    poly1 = Polygon([(1,0),(0,0),(0,1),(1,1)])
    poly2 = Polygon(reversed([(1,1),(1,0),(0,0),(0,1)]))
    assert poly1 != poly2
    
    
def test_equal_polygon3D():
    # type: () -> None
    poly1 = Polygon([(1,0,0),(0,0,0),(0,1,0),(1,1,0)])
    poly2 = Polygon([(1,1,0),(1,0,0),(0,0,0),(0,1,0)])
    assert poly1 == poly2
    
    poly1 = Polygon([(1,0,0),(0,0,0),(0,1,0),(1,1,0)])
    poly2 = Polygon(reversed([(1,1,0),(1,0,0),(0,0,0),(0,1,0)]))
    assert poly1 != poly2
    
    
def test_polygon_index():
    # type: () -> None
    poly = Polygon3D([(0,4,0),(0,0,0),(4,0,0),(4,4,0)])
    assert poly[1] == Vector3D(0,0,0)
    assert poly.index(Vector3D(0,0,0)) == 1


def test_polygon_attributes():
    # type: () -> None
    poly2d = Polygon([(0,0), (0,1), (1,1), (1,0)])
    assert len(poly2d) == 4
    assert poly2d.xs == [0,0,1,1]
    assert poly2d.ys == [0,1,1,0] 
    assert poly2d.zs == [0,0,0,0]
    assert poly2d.vertices_list == [(0,0), (0,1), (1,1), (1,0)]
    assert poly2d.vertices == [Vector2D(*v) for v in poly2d]
    
    
def test_polygon3d_attributes():
    # type: () -> None
    poly3d = Polygon3D([(0,0,0), (0,1,1), (1,1,1), (1,0,0)])
    assert len(poly3d) == 4
    assert poly3d.xs == [0,0,1,1]
    assert poly3d.ys == [0,1,1,0] 
    assert poly3d.zs == [0,1,1,0]
    assert poly3d.vertices_list == [(0,0,0), (0,1,1), (1,1,1), (1,0,0)]
    assert poly3d.vertices == [Vector3D(*v) for v in poly3d]
    assert poly3d.distance == 0
    assert poly3d.is_horizontal == False
    assert poly3d.normal_vector == [0.0, 0.5, -0.5]
    poly3d_2 = Polygon3D([(0,1,1), (0,2,2), (1,2,2), (1,1,1)])
    assert poly3d_2.normal_vector == [0.0, 0.5, -0.5]
    assert poly3d_2.projection_axis == 1
    result = poly3d.is_coplanar(poly3d_2)
    assert result
    
    
def test_add_polygon_to_polygon():
    # type: () -> None
    # 2D
    poly1 = Polygon([(1,0), (0,0), (0,1)])
    poly2 = Polygon([(1,0), (1,0), (1,0)])
    expected = Polygon([(2,0), (1,0), (1,1)])
    result = poly1 + poly2
    assert almostequal(result, expected)

    vector = Vector2D(1,0)    
    result = poly1 + vector
    assert almostequal(result, expected)
    
    vector = Vector3D(1,0,0)
    try:
        result = poly1 + vector  # should fail
        assert False
    except ValueError:
        pass
    # 3D
    poly1 = Polygon([(1,0,1), (0,0,1), (0,1,1)])
    poly2 = Polygon([(1,0,1), (1,0,1), (1,0,1)])
    expected = Polygon([(2,0,2), (1,0,2), (1,1,2)])
    result = poly1 + poly2
    assert almostequal(result, expected)

    vector = Vector3D(1,0,1)    
    result = poly1 + vector
    assert almostequal(result, expected)
    
    vector = Vector2D(1,0)
    try:
        result = poly1 + vector  # should fail
        assert False
    except ValueError:
        pass


def test_polygons_not_equal():
    # type: () -> None
    """Test the check for equality between polygons.
    """
    # different normal vector
    poly3d = Polygon3D([(0,0,0), (0,1,1), (1,1,1), (1,0,0)])
    poly3d_2 = Polygon3D([(0,0,2), (0,1,2), (1,1,2), (1,0,2)])
    assert poly3d.normal_vector != poly3d_2.normal_vector
    assert poly3d != poly3d_2
    # different distance
    poly3d = Polygon3D([(0,0,0), (0,1,1), (1,1,1), (1,0,0)])
    poly3d_2 = Polygon3D([(0,1,2), (0,2,3), (1,2,3), (1,1,2)])
    assert poly3d.normal_vector == poly3d_2.normal_vector
    assert poly3d.distance != poly3d_2.distance
    assert poly3d != poly3d_2
    

def test_order_points():
    # type: () -> None
    polygon = Polygon3D([(0,0,0), (0,1,1), (1,1,1), (1,0,0)])
    starting_position = 'upperleftcorner'
    expected = Polygon3D([(1,1,1), (1,0,0), (0,0,0), (0,1,1)])
    result = polygon.order_points(starting_position)
    assert result == expected
    assert result[0] == expected[0]
    
    starting_position = 'lowerleftcorner'
    expected = Polygon3D([(1,0,0), (0,0,0), (0,1,1), (1,1,1)])
    result = polygon.order_points(starting_position)
    assert result == expected
    assert result[0] == expected[0]
    
    starting_position = 'lowerrightcorner'
    expected = Polygon3D([(0,0,0), (0,1,1), (1,1,1), (1,0,0)])
    result = polygon.order_points(starting_position)
    assert result == expected
    assert result[0] == expected[0]
    
    starting_position = 'upperrightcorner'
    expected = Polygon3D([(0,1,1), (1,1,1), (1,0,0), (0,0,0)])
    result = polygon.order_points(starting_position)
    assert result == expected
    assert result[0] == expected[0]


def test_bounding_box():
    # type: () -> None
    poly = Polygon([(0,0), (0,1), (1,1), (1,0)])
    poly3d = Polygon3D([(0,0,0), (0,1,1), (1,1,1), (1,0,0)])
    
    expected = Polygon([(1,1,1), (1,0,0), (0,0,0), (0,1,1)])

    result = poly3d.bounding_box
    assert almostequal(result, expected)


def test_reflect():
    # type: () -> None
    """
    Test that a polygon with inverted orientation is seen as coplanar with the
    original polygon, but not seen as equal.
    
    """
    poly3d = Polygon3D([(0,0,0), (0,1,1), (1,1,1), (1,0,0)])
    poly3d_inv = poly3d.invert_orientation()
    assert poly3d != poly3d_inv
    assert poly3d.is_coplanar(poly3d_inv)
    

def test_rotate():
    # type: () -> None
    """Test for rotating 3D polygons into 2D and back again
    """
    # At the origin
    s1 = Polygon3D([(0,0,2), (2,0,2), (0,0,0)])  # vertical
    expected = Polygon([(0,2), (2,2), (0,0)])
    # convert to 2D    
    result = s1.project_to_2D()
    assert result == expected
    
    # revert to 3D
    result = result.project_to_3D(s1)
    assert result == s1
    
    # Away from the origin
    s1 = Polygon3D([(1,0,2), (3,0,2), (1,0,0)])  # vertical
    expected = Polygon([(1,2), (3,2), (1,0)])
    # convert to 2D    
    result = s1.project_to_2D()

    assert result == expected

    # revert to 3D
    result = result.project_to_3D(s1)
    assert result == s1

    # Away from the origin
    s1 = Polygon3D([(0,1,1), (2,2,0), (2,2,2), (0,1,2)])  # vertical
    expected = Polygon([(0.0, 1.0), (2.0, 0.0), (2.0, 2.0), (0.0, 2.0)])

    # convert to 2D    
    result = s1.project_to_2D()
    assert result == expected

    # revert to 3D
    result = result.project_to_3D(s1)
    assert result == s1


def test_union_2D_polys_single():
    # type: () -> None
    """Simplest test for union_2D_polys
    
    This has two squares in the horizontal plane which overlap in one place.
    Fails if the expected union shape is not returned.
    
    """
    # surface is already a flat plane with z == 0
    s1 = Polygon([(0,2), (2,2), (2,0), (0,0)])  # clockwise
    s2 = Polygon([(1,3), (3,3), (3,1), (1,1)])  # clockwise
    expected = [Polygon([(0,0), (0,2), (1,2), (1,3),
                                  (3,3), (3,1), (2,1), (2,0)])]  # clockwise
    
    result = union_2D_polys(s1, s2)
    for res, exp in zip(result, expected):
        assert res == exp

    result = s1.union(s2)
    assert res == exp

    result = s2.union(s1)
    assert res == exp


def test_intersect_2D_polys_single():
    # type: () -> None
    """Simplest test for intersect_2D_polys
    
    This has two squares in the horizontal plane which overlap in one place.
    Fails if the expected overlapping shape is not returned.
    
    """
    # surface is already a flat plane with z == 0
    s1 = Polygon([(0,2), (2,2), (2,0), (0,0)])  # clockwise
    s2 = Polygon([(1,3), (3,3), (3,1), (1,1)])  # clockwise
    expected = [Polygon([(1,2), (2,2), (2,1), (1,1)])]  #clockwise
    
    result = s1.intersect(s2)
    for res, exp in zip(result, expected):
        assert res == exp

    result = s2.intersect(s1)
    for res, exp in zip(result, expected):
        assert res == exp


def test_difference_2D_polys_single():
    # type: () -> None
    """Simplest test for difference_2D_polys
    
    This has two squares in the horizontal plane which overlap in one place.
    Fails if the two original polygons do not have the intersection removed.
    
    """
    s1 = Polygon([(0,2), (2,2), (2,0), (0,0)])  # clockwise
    s2 = Polygon([(1,3), (3,3), (3,1), (1,1)])  # clockwise
    
    # clockwise
    ex_s1 = [Polygon([(0,2), (1,2), (1,1), (2,1), (2,0), (0,0)])]
    ex_s2 = [Polygon([(1,3), (3,3), (3,1), (2,1), (2,2), (1,2)])]
    expected = [ex_s1, ex_s2]

    result = [s1.difference(s2), s2.difference(s1)]
    assert result[0] == expected[0]
    assert result[1] == expected[1]


def test_union_3D_polys_single():
    # type: () -> None
    """Simplest test for union_3D_polys
    
    This has two squares in the horizontal plane which overlap in one place.
    Fails if the expected union shape is not returned.
    
    """
    # surface is already a flat plane with z == 0
    s1 = Polygon3D([(0,2,0), (2,2,0), (2,0,0), (0,0,0)])  # clockwise
    s2 = Polygon3D([(1,3,0), (3,3,0), (3,1,0), (1,1,0)])  # clockwise
    expected = [Polygon3D([(0,0,0), (0,2,0), (1,2,0), (1,3,0),
                           (3,3,0), (3,1,0), (2,1,0), (2,0,0)])]  # clockwise
    
    result = union_3D_polys(s1, s2)
    for res, exp in zip(result, expected):
        assert res == exp

    result = s1.union(s2)
    assert res == exp

    result = s2.union(s1)
    assert res == exp


def test_intersect_3D_polys_single():
    # type: () -> None
    """Simplest test for intersect_3D_polys
    
    This has two squares in the horizontal plane which overlap in one place.
    Fails if the expected overlapping shape is not returned.
    
    """
    # surface is already a flat plane with z == 0
    s1 = Polygon3D([(0,2,0), (2,2,0), (2,0,0), (0,0,0)])  # clockwise
    s2 = Polygon3D([(1,3,0), (3,3,0), (3,1,0), (1,1,0)])  # clockwise
    expected = [Polygon3D([(1,2,0), (2,2,0), (2,1,0), (1,1,0)])]  #clockwise
    result = intersect_3D_polys(s1, s2)
    for res, exp in zip(result, expected):
        assert res == exp


def test_difference_3D_polys_single():
    # type: () -> None
    """Simplest test for difference_3D_polys
    
    This has two squares in the horizontal plane which overlap in one place.
    Fails if the two original polygons do not have the intersection removed.
    
    """
    # surface is already a flat plane with z == 0
    s1 = Polygon3D([(0,2,0), (2,2,0), (2,0,0), (0,0,0)])  # clockwise
    s2 = Polygon3D([(1,3,0), (3,3,0), (3,1,0), (1,1,0)])  # clockwise
    
    # clockwise
    ex_s1 = [Polygon3D([(0,2,0), (1,2,0), (1,1,0), (2,1,0), (2,0,0), (0,0,0)])]
    ex_s2 = [Polygon3D([(1,3,0), (3,3,0), (3,1,0), (2,1,0), (2,2,0), (1,2,0)])]
    expected = [ex_s1, ex_s2]

    result = [difference_3D_polys(s1, s2), difference_3D_polys(s2, s1)]
    assert result[0] == expected[0]
    assert result[1] == expected[1]
    
    
def test_intersect_3D_polys_angled():
    # type: () -> None
    s1 = Polygon3D([(2.5,1.95,0.5), (2.5,1.95,0), (1.5,2.05,0), (1.5,2.05,0.5)])  # clockwise
    s2 = Polygon3D([(1,2.1,0.5), (1,2.1,0), (2,2,0), (2,2,0.5)])  # clockwise
    expected = [Polygon3D([(2.0, 2.0, 0.0), (1.5, 2.05, 0.0),
                           (1.5, 2.05, 0.5),(2.0, 2.0, 0.5)])]

    result = intersect_3D_polys(s1, s2)

    assert result == expected


def test_intersect_no_overlap():
    # type: () -> None
    # surfaces don't overlap
    s1 = Polygon3D([(0,2,0), (2,2,0), (2,0,0), (0,0,0)])  # clockwise
    s2 = Polygon3D([(2,3,0), (3,3,0), (3,1,0), (2,1,0)])  # clockwise
    expected = []  #clockwise
    result = intersect_3D_polys(s1, s2)
    assert result == expected


def test_difference_no_difference():
    # type: () -> None
    # surfaces don't overlap
    s1 = Polygon3D([(0,2,0), (2,2,0), (2,0,0), (0,0,0)])  # clockwise
    s2 = s1
    expected = []  #clockwise
    result = difference_3D_polys(s1, s2)
    assert result == expected


def test_intersect_3D_polys_multi():
    # type: () -> None
    """Test for intersect_3D_polys with two overlapping regions
    
    This has two shapes in the horizontal plane which overlap in two places.
    Fails if the overlapping shapes are not returned as a new polygons.

    """    
    # surface is already a flat plane with z == 0
    s1 = Polygon3D([(0,0,0), (5,0,0), (5,2,0), (0,2,0)])  # counterclockwise
    s2 = Polygon3D([(1,1,0), (2,1,0), (2,2,0), (3,2,0),
                     (3,1,0), (4,1,0), (4,3,0), (1,3,0)])  # counterclockwise
    overlap = [Polygon3D([(1,1,0), (2,1,0), (2,2,0), (1,2,0)]),
               Polygon3D([(3,1,0), (4,1,0), (4,2,0), (3,2,0)])]
    
    ex_s1 = Polygon3D([(0,0,0), (5,0,0), (5,2,0), (4,2,0), (4,1,0), (3,1,0), 
                        (3,2,0), (2,2,0), (2,1,0), (1,1,0), (1,2,0), (0,2,0)])
    ex_s2 = Polygon3D([(1,2,0), (4,2,0), (4,3,0), (1,3,0)])

    expected = [ex_s1, ex_s2]
    expected.extend(overlap)
    result = intersect_3D_polys(s1, s2)
    
    for res, exp in zip(result, overlap):
        assert res == exp

    result = s1.intersect(s2)
    for res, exp in zip(result, overlap):
        assert res == exp

    result = s2.intersect(s1)
    for res, exp in zip(result, overlap):
        assert res == exp


def test_difference_3D_polys_multi():
    # type: () -> None
    """Test for difference_3D_polys with two overlapping regions
    
    This has two shapes in the horizontal plane which overlap in two places.
    Fails if the overlapping shapes are not returned as a new polygons.

    """    
    # surface is already a flat plane with z == 0
    s1 = Polygon3D([(0,0,0), (5,0,0), (5,2,0), (0,2,0)])  # counterclockwise
    s2 = Polygon3D([(1,1,0), (2,1,0), (2,2,0), (3,2,0),
                     (3,1,0), (4,1,0), (4,3,0), (1,3,0)])  # counterclockwise
    
    ex_s1 = [Polygon3D([(0,0,0), (5,0,0), (5,2,0), (4,2,0), (4,1,0), (3,1,0), 
                        (3,2,0), (2,2,0), (2,1,0), (1,1,0), (1,2,0), (0,2,0)])]
    ex_s2 = [Polygon3D([(1,2,0), (4,2,0), (4,3,0), (1,3,0)])]

    expected = [ex_s1, ex_s2]
    result = [difference_3D_polys(s1, s2), difference_3D_polys(s2, s1)]
    
    for res, exp in zip(result, expected):
        assert len(res) == len(exp)
        for r, e  in zip(res, exp):
            assert r == e
            
    assert s1.difference(s2) == ex_s1
    assert s2.difference(s1) == ex_s2


def test_surface_normal():
    # type: () -> None
    poly = Polygon3D([Vector3D(0.0, 0.0, 0.0),
                      Vector3D(1.0, 0.0, 0.0),
                      Vector3D(1.0, 1.0, 0.0),
                      Vector3D(0.0, 1.0, 0.0)])
    assert list(poly.normal_vector) == [0.0, 0.0, 1.0]  # for a horizontal surface

    poly = Polygon3D(reversed([Vector3D(0.0, 0.0, 0.0),
                      Vector3D(1.0, 0.0, 0.0),
                      Vector3D(1.0, 1.0, 0.0),
                      Vector3D(0.0, 1.0, 0.0)]))
    assert list(poly.normal_vector) == [0.0, 0.0, -1.0]  # for a horizontal surface

    poly = Polygon3D([Vector3D(0.0, 0.0, 0.0),
                      Vector3D(2.0, 1.0, 0.0),
                      Vector3D(4.0, 0.0, 0.0),
                      Vector3D(4.0, 3.0, 0.0),
                      Vector3D(2.0, 2.0, 0.0),
                      Vector3D(0.0, 3.0, 0.0)])
    assert list(poly.normal_vector) == [0.0, 0.0, 1.0]  # for a horizontal surface

    poly = Polygon3D(reversed([Vector3D(0.0, 0.0, 0.0),
                      Vector3D(2.0, 1.0, 0.0),
                      Vector3D(4.0, 0.0, 0.0),
                      Vector3D(4.0, 3.0, 0.0),
                      Vector3D(2.0, 2.0, 0.0),
                      Vector3D(0.0, 3.0, 0.0)]))
    assert list(poly.normal_vector) == [0.0, 0.0, -1.0]  # for a horizontal surface
    
    poly = Polygon3D([[ 1.,  1.1,  0.5],
                      [ 1.,  1.1,  0.],
                      [ 1.,  2.1,  0.],
                      [ 1.,  2.1,  0.5]])
    assert list(poly.normal_vector) == [1.0, 0.0, 0.0]  # for a horizontal surface

    
def test_surface_is_clockwise():
    # type: () -> None
    """Test if a surface is clockwise as seen from a given point.
    """
    poly = Polygon3D(reversed([
        Vector3D(0.0, 0.0, 0.0),
        Vector3D(1.0, 0.0, 0.0),
        Vector3D(1.0, 1.0, 0.0),
        Vector3D(0.0, 1.0, 0.0)]))
    poly_inv = Polygon3D([
        Vector3D(0.0, 0.0, 0.0),
        Vector3D(1.0, 0.0, 0.0),
        Vector3D(1.0, 1.0, 0.0),
        Vector3D(0.0, 1.0, 0.0)])

    pt = Vector3D(0.5, 0.5, 1.0)  # point above the plane

    assert poly.is_clockwise(pt)
    assert not poly_inv.is_clockwise(pt)
    

def test_break_polygons():
    # type: () -> None
    poly = Polygon3D([(0,4,0),(0,0,0),(4,0,0),(4,4,0)])
    hole = Polygon3D([(1,3,0),(1.5,2,0),(1,1,0),(3,1,0),(3,3,0)])
    expected = [
        Polygon3D([(0,4,0),(0,0,0),(1,1,0),(1.5,2,0),(1,3,0)]),
        Polygon3D([(4,4,0),(0,4,0),(1,3,0),(3,3,0),(3,1,0),(1,1,0),(0,0,0),(4,0,0)])
        ]

    result = break_polygons(poly, hole)
    
    assert result[0] == expected[0]
    assert result[1] == expected[1]


def test_identify_inner_ring_polygons():
    # type: () -> None
    geom = [
        (u'POLYGON((528318.562 186087.89,528307.65 186095.1,528303.1 186088.15,528305.59 186086.5,528299.67 186077.17,528300.85 186075.56,528295.35 186071.9,528294.401 186072.514,528292.8 186073.55,528285.8 186069.2,528286 186066.5,528286.45 186065.85,528285.854 186065.438,528277.752 186059.839,528275.54 186058.31,528273.93 186066.89,528273.76 186067.81,528273.45 186069.42,528269.73 186068.72,528267.91 186078.41,528242.51 186073.65,528242.78 186072.209,528243.49 186068.42,528252.68 186058.13,528255.36 186043.87,528261.22 186044.97,528266.75 186016.45,528269.25 186016.95,528271 186017.3,528271.2 186016.25,528277.85 186017.5,528277.25 186020.4,528300.75 186024.85,528301 186023.65,528307.8 186024.95,528307.7 186025.6,528311.5 186028.5,528310.85 186029.4,528316.7 186033.3,528320.15 186035.55,528320.7 186034.95,528324.15 186037.25,528325 186036.95,528328.75 186042.7,528327.85 186043.3,528340.6 186063.2,528342.65 186061.8,528347.25 186068.8,528347.3 186068.9,528318.562 186087.89),(528306.281 186071.421,528312.36 186075.46,528314.06 186078.02,528314.7 186077.6,528315.98 186076.78,528311.35 186069.56,528320.79 186063.52,528317.54 186058.45,528318.74 186057.68,528315.04 186051.9,528318.42 186049.74,528317.5 186048.31,528319.21 186047.21,528317.9 186045.2,528316.6 186043.15,528302.35 186063.6,528304.4 186065.05,528301.99 186068.57,528306.281 186071.421),(528294.9 186046.92,528296.47 186044.6,528299.04 186040.78,528298.55 186040.45,528290.62 186038.77,528289.3 186038.48,528288.33 186043.03,528289.65 186043.28,528287.66 186052.57,528285.16 186052.04,528284.85 186053.5,528284.1 186054.8,528286.79 186056.63,528287.12 186056.86,528286.6 186057.65,528287.3 186058.14,528289.45 186055,528291.3 186056.25,528296.85 186048.23,528294.9 186046.92),(528274.89 186044.69,528277.99 186045.28,528276.9 186051.08,528276.98 186051.1,528278.3 186052,528280.4 186053.44,528280.72 186052.98,528281.04 186051.33,528278.53 186050.86,528280.42 186041.51,528282.23 186041.86,528283.17 186037.18,528276.65 186035.8,528274.89 186044.69))', 7.7)]
    
    for item in geom:
        poly = Polygon3D([]).from_wkt(item[0])
#    view_polygons({'blue': [poly]})#, 'red': [interior]})
    

def test_point():
    # type: () -> None
    pt1 = Vector3D(0.0, 0.0, 0.0)
    pt2 = Vector3D(1.0, 1.0, 1.0)
    
    assert pt2 - pt1 == pt2
    assert pt1 - pt2 == Vector3D(-1,-1,-1)
    
    assert pt2 + pt2 == Vector3D(2,2,2)
           
           
def test_invert():
    # type: () -> None
    v = Vector3D(1, 2, 3)
    assert v.invert() == Vector3D(-1, -2, -3)
    

def test_set_length():
    # type: () -> None
    v = Vector3D(1, 1, 1)
    v.set_length(1)
    for i in v:
        assert almostequal(i, 0.57735026)


def test_normalize():
    # type: () -> None
    v = Vector3D(1, 1, 1)
    v.normalize()
    for i in v:
        assert almostequal(i, 0.57735026)


def test_on_poly_edge():
    # type: () -> None
    poly = Polygon3D([(0,4,0),(0,0,0),(4,0,0),(4,4,0)])
    edge1 = Segment(Vector3D(0,1,0), Vector3D(0,2,0))
    edge2 = Segment(Vector3D(1,1,0), Vector3D(1,2,0))
    assert edge1._on_poly_edge(poly)
    assert not edge2._on_poly_edge(poly)


def test_closest():
    # type: () -> None
    pt = Vector3D(0,0,0)
    poly = Polygon3D([(1,1,1), (2,2,3), (3,4,5)])
    assert pt.closest(poly) == Vector3D(1,1,1)




# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""pytest for builder.py"""
from eppy.iddcurrent import iddcurrent
from six import StringIO

from geomeppy.eppy_patches import IDF
from tests.test_builder import idf_txt


idf_txt = """
Version, 8.5;
"""


class TestAddShadingBlock():

    def setup(self):
        # type: () -> None
        iddfhandle = StringIO(iddcurrent.iddtxt)
        if IDF.getiddname() == None:
            IDF.setiddname(iddfhandle)
        
        self.idf = IDF(StringIO(idf_txt))

    def test_add_shading_block_smoke_test(self):
        # type: () -> None
        idf = self.idf
        name = "test"
        height = 7.5
        coordinates = [
            (87.25,24.0),(91.7,25.75),(90.05,30.25),
            (89.55,31.55),(89.15,31.35),(85.1,29.8),
            (86.1,27.2),(84.6,26.65),(85.8,23.5),
            (87.25,24.0)]
        idf.add_shading_block(name, coordinates, height)
        
        



# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""Tool for visualising geometry."""
from eppy.function_helpers import getcoords
from eppy.iddcurrent import iddcurrent
from eppy.modeleditor import IDF
from six import StringIO

try:
    from mpl_toolkits.mplot3d import Axes3D  # noqa
    from mpl_toolkits.mplot3d.art3d import Poly3DCollection
    import matplotlib.pyplot as plt
except ImportError:
    # this isn't always needed so we can ignore if it's not present
    pass


def view_idf(fname=None, idf_txt=None):
    # type: (Optional[str], Optional[str]) -> None
    """Display an IDF for inspection.

    :param fname: Path to the IDF.
    :param idf_txt: The string representation of an IDF.
    """
    if fname and idf_txt:
        raise ValueError('Pass either fname or idf_txt, not both.')
    # set the IDD for the version of EnergyPlus
    iddfhandle = StringIO(iddcurrent.iddtxt)
    if IDF.getiddname() is None:
        IDF.setiddname(iddfhandle)

    if fname:
        # import the IDF
        idf = IDF(fname)
    elif idf_txt:
        idf = IDF()
        idf.initreadtxt(idf_txt)
    # create the figure and add the surfaces
    plt.figure()
    ax = plt.axes(projection='3d')
    collections = _get_collections(idf, opacity=0.5)
    for c in collections:
        ax.add_collection3d(c)

    # calculate and set the axis limits
    limits = _get_limits(idf=idf)
    ax.set_xlim(limits['x'])
    ax.set_ylim(limits['y'])
    ax.set_zlim(limits['z'])

    plt.show()


def view_polygons(polygons):
    """Display a collection of polygons for inspection.

    :param polygons: A dict keyed by colour, containing Polygon3D objects to show in that colour.
    """
    # create the figure and add the surfaces
    plt.figure()
    ax = plt.axes(projection='3d')

    collections = _make_collections(polygons, opacity=0.5)

    for c in collections:
        ax.add_collection3d(c)

    # calculate and set the axis limits
    limits = _get_limits(polygons=polygons)
    ax.set_xlim(limits['x'])
    ax.set_ylim(limits['y'])
    ax.set_zlim(limits['z'])

    plt.show()


def _get_surfaces(idf):
    """Get the surfaces from the IDF.
    """
    surface_types = ['BUILDINGSURFACE:DETAILED',
                     'FENESTRATIONSURFACE:DETAILED']
    surfaces = []
    for surface_type in surface_types:
        surfaces.extend(idf.idfobjects[surface_type])

    return surfaces


def _get_collections(idf, opacity=1):
    """Set up 3D collections for each surface type.
    """
    surfaces = _get_surfaces(idf)
    # set up the collections
    walls = Poly3DCollection([getcoords(s) for s in surfaces
                              if s.Surface_Type.lower() == 'wall'],
                             alpha=opacity,
                             facecolor='wheat',
                             edgecolors='black'
                             )
    floors = Poly3DCollection([getcoords(s) for s in surfaces
                               if s.Surface_Type.lower() == 'floor'],
                              alpha=opacity,
                              facecolor='dimgray',
                              edgecolors='black'
                              )
    roofs = Poly3DCollection([getcoords(s) for s in surfaces
                              if s.Surface_Type.lower() == 'roof'],
                             alpha=opacity,
                             facecolor='firebrick',
                             edgecolors='black'
                             )
    windows = Poly3DCollection([getcoords(s) for s in surfaces
                                if s.Surface_Type.lower() == 'window'],
                               alpha=opacity,
                               facecolor='cornflowerblue',
                               edgecolors='black'
                               )

    return walls, roofs, floors, windows


def _make_collections(polygons, opacity=1):
    """Make collections from a dict of polygons.
    """
    collection = []
    for color in polygons:
        collection.append(Poly3DCollection(
            [p.points_matrix for p in polygons[color]],
            alpha=opacity,
            facecolor=color,
            edgecolors='black'))
    return collection


def _get_limits(idf=None, polygons=None):
    """Get limits for the x, y and z axes so the plot is fitted to the axes."""
    if polygons:
        x = [pt[0] for color in polygons for p in polygons[color] for pt in p]
        y = [pt[1] for color in polygons for p in polygons[color] for pt in p]
        z = [pt[2] for color in polygons for p in polygons[color] for pt in p]

    elif idf:
        surfaces = _get_surfaces(idf)

        x = [pt[0] for s in surfaces for pt in getcoords(s)]
        y = [pt[1] for s in surfaces for pt in getcoords(s)]
        z = [pt[2] for s in surfaces for pt in getcoords(s)]

    max_delta = max((max(x) - min(x)),
                    (max(y) - min(y)),
                    (max(z) - min(z)))

    return {'x': (min(x), min(x) + max_delta),
            'y': (min(y), min(y) + max_delta),
            'z': (min(z), min(y) + max_delta)}


def main(fname=None, polygons=None):
    if fname:
        view_idf(fname)
    elif polygons:
        view_polygons(polygons)


if __name__ == "__main__":
    main()



import os

from setuptools import setup


THIS_DIR = os.path.dirname(os.path.abspath(__file__))

def read_md(f):
    try:
        from pypandoc import convert
        try:
            return convert(os.path.join(THIS_DIR, f), 'rst')
        except:
            return "GeomEppy"
    except ImportError:
        print("warning: pypandoc module not found, could not convert Markdown to RST")
        try:
            with open(os.path.join(THIS_DIR, f), 'r') as f_in:
                return f_in.read()
        except:
            return "GeomEppy"

setup(
    name='geomeppy',
    packages=['geomeppy',
              'tests',
              ],
    version='0.3.9',
    description='Geometry editing for E+ idf files',
    long_description=read_md('README.md'),
    author='Jamie Bull',
    author_email='jamie.bull@oco-carbon.com',
    url='https://github.com/jamiebull1/geomeppy',
    download_url='https://github.com/jamiebull1/geomeppy/tarball/v0.3.7',
    license='MIT License',
    keywords=['EnergyPlus',
              'geometry',
              'building performance simulation',
              ],
    platforms='any',
    install_requires=[
        'eppy==0.5.46',
        'numpy==1.13',
        'pyclipper==1.0.6',  # geometry intersection
        'shapely==1.6.2',  # geometry transformations
        'six==1.11.0',  # python2/3 compatibility
        'transforms3d==0.3.1',  # geometry transformations
        ],
    classifiers = [
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Development Status :: 3 - Alpha',
        'Natural Language :: English',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering',
        ],
    extras_require={
        ':python_version>="3.4"': [
            'mypy==0.550',  # static type checking
        ],
        'testing': [
            'codecov',
            'flake8',
            'pytest-cov',
            'typing',
        ],
    }
)

